import queue

import pytest

from .notifier import Notifier
from .test_helper import get_exception, build_notice_from_str


def test_build_notice_from_exception():
  notifier = Notifier()

  err = get_exception()
  notice = notifier.build_notice(err)

  errors = notice['errors']
  assert len(errors) == 1

  error = errors[0]
  assert error['type'] == 'ValueError'
  assert error['message'] == 'hello'

  backtrace = error['backtrace']
  assert len(backtrace) == 1

  frame = backtrace[0]
  assert frame['file'] == '[PROJECT_ROOT]/pybrake/test_helper.py'
  assert frame['function'] == 'get_exception'
  assert frame['line'] == 3
  print(frame['code'])
  assert frame['code'] == {
    1: 'def get_exception():',
    2: '  try:',
    3: "    raise ValueError('hello')",
    4: '  except Exception as err:',
    5: '    return err',
  }

  context = notice['context']
  assert context['notifier'] == {
    'name': 'pybrake',
    'url': 'https://github.com/airbrake/pybrake',
    'version': '0.0.1',
  }
  for k in ['os', 'language', 'hostname']:
    assert context[k]


def test_build_notice_from_str():
  notifier = Notifier()

  notice = build_notice_from_str(notifier, 'hello')

  errors = notice['errors']
  assert len(errors) == 1

  error = errors[0]
  assert error['message'] == 'hello'

  backtrace = error['backtrace']
  assert len(backtrace) >= 1

  frame = backtrace[0]
  assert frame['file'] == '[PROJECT_ROOT]/pybrake/test_helper.py'
  assert frame['function'] == 'build_notice_from_str'
  assert frame['line'] == 9
  assert frame['code'] == {
    7: '',
    8: 'def build_notice_from_str(notifier, s):',
    9: '    return notifier.build_notice(s)',
    10: '',
    11: '',
  }


def test_environment():
  notifier = Notifier(environment='production')

  notice = notifier.build_notice('hello')

  assert notice['context']['environment'] == 'production'


def test_root_directory():
  notifier = Notifier(root_directory='/root/dir')

  notice = notifier.build_notice('hello')

  assert notice['context']['rootDirectory'] == '/root/dir'


def test_filter_data():
  def filter(notice):
    notice['params']['param'] = 'value'
    return notice

  notifier = Notifier()
  notifier.add_filter(filter)

  notice = notifier.notify_sync('hello')

  assert notice['params']['param'] == 'value'


def test_filter_ignore():
  notifier = Notifier()
  notifier.add_filter(lambda notice: None)

  notice = notifier.notify_sync('hello')

  assert notice['error'] == 'notice is filtered out'


def test_filter_ignore_async():
  notifier = Notifier()
  notifier.add_filter(lambda notice: None)

  future = notifier.notify('hello')
  notice = future.result()

  assert notice['error'] == 'notice is filtered out'


def test_unauthorized():
  notifier = Notifier()

  notice = notifier.notify_sync('hello')

  assert notice['error'] == 'API key is required'


def test_unknown_host():
  notifier = Notifier(host='http://airbrake123.com')

  notice = notifier.notify_sync('hello')

  assert str(notice['error']) == '<urlopen error [Errno -2] Name or service not known>'


def test_truncation():
  notifier = Notifier()

  notice = notifier.build_notice('hello')
  notice['params']['param'] = 'x' * 64000
  notice = notifier.send_notice_sync(notice)

  assert len(notice['params']['param']) == 1024


def _test_full_queue():
  notifier = Notifier(max_queue_size=10)

  for i in range(100):
    future = notifier.notify('hello')

  notifier.close()

  notice = future.result()
  assert notice['error'] == 'queue is full'


def _test_rate_limited():
  notifier = Notifier()

  for i in range(101):
    future = notifier.notify('hello')

  notice = future.result()
  assert notice['error'] == 'IP is rate limited'



import json
import collections
import collections.abc


_MAX_JSON_SIZE = 64000 # 64kb
_NOTICE_KEYS = ['context', 'params', 'session', 'environment']


def jsonify_notice(notice):
  b = _jsonify(notice)
  if len(b) < _MAX_JSON_SIZE:
    return b.encode('utf8')

  if 'errors' in notice:
    t = Truncator(0)
    for i, error in enumerate(notice['errors']):
      notice['errors'][i] = t.truncate_dict(error)

  for level in range(10):
    b = _jsonify(notice)
    if len(b) < _MAX_JSON_SIZE:
      break

    t = Truncator(level)
    for key in _NOTICE_KEYS:
      if key in notice:
        notice[key] = t.truncate_dict(notice[key])

  return b.encode('utf8')


def _jsonify(obj):
  return json.dumps(obj, default=set_default)


def set_default(obj):
  if isinstance(obj, set):
    return list(obj)
  if isinstance(obj, (bytes, bytearray)):
    return obj.decode('utf8', 'replace')
  if isinstance(obj, (collections.UserDict,
                      collections.UserList,
                      collections.UserString)):
    return obj.data
  return str(obj)


class Truncator:
  def __init__(self, level=0):
    self._max_dict_len = _scale(128, level)
    self._max_list_len = _scale(128, level)
    self._max_str_len = _scale(1024, level)

  def truncate_dict(self, d):
    res = dict()
    i = 0
    for k, v in d.items():
      res[k] = self.truncate(v)
      i += 1
      if i >= self._max_dict_len:
        break
    return res

  def truncate_list(self, l):
    res = list()
    for i, v in enumerate(l):
      res.append(self.truncate(v))
      if i >= self._max_list_len:
        break
    return res

  def truncate_str(self, s):
    if len(s) <= self._max_str_len:
      return s
    return s[:self._max_str_len]

  def truncate(self, v):
    if isinstance(v, str):
      return self.truncate_str(v)
    if isinstance(v, (bytes, bytearray)):
      return self.truncate_str(v.decode('utf8', 'replace'))
    if isinstance(v, collections.UserString):
      return self.truncate_str(v.data)

    if isinstance(v, collections.abc.Mapping):
      return self.truncate_dict(v)

    if isinstance(v, collections.abc.Iterable):
      return self.truncate_list(v)

    return v


def _scale(num, level):
  return (num >> level) or 1



import os
import sys
import platform
import traceback as tb
import socket
import urllib.request, urllib.error
import queue
from concurrent import futures
import json
import time
import logging

from .notice import jsonify_notice
from .code_hunks import get_code_hunk
from .utils import get_logger


_ERR_IP_RATE_LIMITED = 'IP is rate limited'

_AIRBRAKE_URL_FORMAT = '{}/api/v3/projects/{}/notices'

_CONTEXT = dict(
  notifier=dict(
    name='pybrake',
    version='0.0.3',
    url='https://github.com/airbrake/pybrake',
  ),
  os=platform.platform(),
  language='Python/%s' % platform.python_version(),
  hostname=socket.gethostname(),
  versions=dict(
    python=platform.python_version(),
  ),
)

_logger = get_logger()


class Notifier:
  def __init__(self,
               project_id=0,
               project_key='',
               host='https://api.airbrake.io',
               **kwargs):
    self._filters = []
    self._rate_limit_reset = 0
    self._max_queue_size = kwargs.get('max_queue_size', 1000)
    self._thread_pool = None

    self._airbrake_url = _AIRBRAKE_URL_FORMAT.format(host, project_id)
    self._airbrake_headers = {'authorization': 'Bearer ' + project_key,
                              'content-type': 'application/json'}

    self._context = _CONTEXT.copy()
    self._context['rootDirectory'] = kwargs.get('root_directory', os.getcwd())
    if 'environment' in kwargs:
      self._context['environment'] = kwargs['environment']

  def close(self):
    if self._thread_pool is not None:
      self._thread_pool.shutdown()

  def add_filter(self, filter_fn):
    """Appends filter to the list.

    Filter is a function that accepts notice. Filter can modify passed
    notice or return None if notice must be ignored.
    """
    self._filters.append(filter_fn)

  def notify_sync(self, err):
    """Notifies Airbrake about exception.

    Under the hood notify is a shortcut for build_notice and send_notice.
    """
    notice = self.build_notice(err)
    return self.send_notice_sync(notice)

  def build_notice(self, err):
    """Builds Airbrake notice from the exception."""
    notice = dict(
      errors=self._build_errors(err),
      context=self._build_context(),
      params=dict(
        sys_executable=sys.executable,
        sys_path=sys.path,
      ),
    )
    return notice

  def send_notice_sync(self, notice):
    """Sends notice to Airbrake.

    It returns notice with 2 possible new keys:
    - {'id' => str} - notice id on success.
    - {'error' => str|Exception} - error on failure.
    """
    for fn in self._filters:
      r = fn(notice)
      if r is None:
        notice['error'] = 'notice is filtered out'
        return notice
      notice = r

    if time.time() < self._rate_limit_reset:
      notice['error'] = _ERR_IP_RATE_LIMITED
      return notice

    data = jsonify_notice(notice)
    req = urllib.request.Request(self._airbrake_url,
                                 data=data,
                                 headers=self._airbrake_headers)

    try:
      resp = urllib.request.urlopen(req, timeout=5)
    except urllib.error.HTTPError as err:
      resp = err
    except Exception as err:
      notice['error'] = err
      _logger.error(notice['error'])
      return notice

    try:
      body = resp.read()
    except Exception as err:
      notice['error'] = err
      _logger.error(notice['error'])
      return notice

    if not (200 <= resp.code < 300 or 400 <= resp.code < 500):
      notice['error'] = 'unexpected Airbrake response status code'
      notice['error_info'] = dict(code=resp.code, body=body)
      _logger.error(notice['error'])
      return notice

    if resp.code == 429:
      return self._rate_limited(notice, resp)

    try:
      data = json.loads(body.decode('utf-8'))
    except Exception as err:
      notice['error'] = err
      _logger.error(notice['error'])
      return notice

    if 'id' in data:
      notice['id'] = data['id']
      return notice

    if 'message' in data:
      notice['error'] = data['message']
      _logger.error(notice['error'])
      return notice

    notice['error'] = 'unexpected Airbrake response'
    notice['error_info'] = dict(data=data)
    _logger.error(notice['error'])
    return notice

  def _rate_limited(self, notice, resp):
    v = resp.headers.get('X-RateLimit-Delay')
    if v is None:
      notice['error'] = 'X-RateLimit-Delay header is missing'
      return notice

    try:
      delay = int(v)
    except ValueError as err:
      notice['error'] = err
      return notice

    self._rate_limit_reset = time.time() + delay

    notice['error'] = _ERR_IP_RATE_LIMITED
    return notice;

  def notify(self, err):
    """Asynchronously notifies Airbrake about exception from separate thread.

    Returns concurrent.futures.Future.
    """
    notice = self.build_notice(err)
    return self.send_notice(notice)

  def send_notice(self, notice):
    """Asynchronously sends notice to Airbrake from separate thread.

    Returns concurrent.futures.Future.
    """
    pool = self._get_thread_pool()
    if pool._work_queue.qsize() >= self._max_queue_size:
      notice['error'] = 'queue is full'
      f = futures.Future()
      f.set_result(notice)
      return f
    return pool.submit(self.send_notice_sync, notice)

  def _build_errors(self, err):
    if isinstance(err, str):
      backtrace = self._build_backtrace_frame(sys._getframe())
      return [{
        'message': err,
        'backtrace': backtrace,
      }]

    errors = []
    while err != None:
      errors.append(self._build_error(err))
      err = err.__cause__ or err.__context__
    return errors

  def _build_error(self, err):
    backtrace = self._build_backtrace_tb(err.__traceback__)
    error = dict(
      type=err.__class__.__name__,
      message=str(err),
      backtrace=backtrace,
    )
    return error

  def _build_backtrace_tb(self, tb):
    backtrace = []
    for frame, lineno in self._walk_tb(tb):
      f = self._build_frame(frame, lineno)
      if f:
        backtrace.append(f)
    return backtrace

  def _walk_tb(self, tb):
    while tb is not None:
      yield tb.tb_frame, tb.tb_lineno
      tb = tb.tb_next

  def _build_backtrace_frame(self, f):
    backtrace = []
    for frame, lineno in self._walk_stack(f):
      f = self._build_frame(frame, lineno)
      if f:
        backtrace.append(f)
    return backtrace

  def _walk_stack(self, f):
    while f is not None:
      yield f, f.f_lineno
      f = f.f_back

  def _build_frame(self, f, line):
    if f.f_locals.get('__traceback_hide__'):
      return None

    filename = f.f_code.co_filename
    func = f.f_code.co_name

    if filename.endswith('pybrake/notifier.py'):
      return None

    loader = f.f_globals.get('__loader__')
    module_name = f.f_globals.get('__name__')
    return self._frame_with_code(filename, func, line,
                                 loader=loader,
                                 module_name=module_name)

  def _frame_with_code(self, filename, func, line, loader=None, module_name=None):
    frame = dict(
      file=self._clean_filename(filename),
      function=func,
      line=line,
    )

    lines = get_code_hunk(filename, line, loader=loader, module_name=module_name)
    if lines is not None:
      frame['code'] = lines

    return frame

  def _clean_filename(self, s):
    if '/lib/python' in s and '/site-packages/' in s:
      needed = '/site-packages/'
      ind = s.find(needed)
      if ind > -1:
        s = '[SITE_PACKAGES]/' + s[ind+len(needed):]
        return s

    s = s.replace(self._context['rootDirectory'], '[PROJECT_ROOT]')
    return s

  def _build_context(self):
    ctx = self._context.copy()
    return ctx

  def _get_thread_pool(self):
    if self._thread_pool is None:
      max_workers = (os.cpu_count() or 1) * 5
      self._thread_pool = futures.ThreadPoolExecutor(max_workers=max_workers)
    return self._thread_pool



import collections
import pytest

from .notice import jsonify_notice


def make_defaultdict():
  d = collections.defaultdict()
  for i in range(100000):
    d[i] = i
  return d


@pytest.mark.parametrize('param, wanted_size', [
  (list(range(100000)), 558),
  (set(range(100000)), 558),
  ({i:i for i in range(100000)}, 1339),
  ('x' * 100000, 1049),
  (b'x' * 100000, 1049),
  (bytearray([64] * 100000), 1049),
  (bytearray([64] * 100000 + [128]), 1049), # UnicodeDecodeError
  (make_defaultdict(), 1339),
  (collections.OrderedDict({i:i for i in range(100000)}), 1339),
  (collections.UserDict({i:i for i in range(100000)}), 1339),
  (collections.UserList(range(100000)), 558),
  (collections.UserString('x' * 100000), 1049),
  (collections.UserString(b'x' * 100000), 1049),
])
def test_truncation(param, wanted_size):
  notice = dict(
    params=dict(param=param),
  )

  b = jsonify_notice(notice)
  assert len(b) == wanted_size



from setuptools import setup


def readme():
  with open('README.rst') as f:
    return f.read()

setup(name='pybrake',
      version='0.1.3',
      description='Python exception notifier for Airbrake',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
      ],
      keywords='airbrake exception error notifier',
      url='http://github.com/airbrake/pybrake',
      author='Vladimir Mihailenco',
      author_email='vlad@airbrake.io',
      license='MIT',
      packages=['pybrake'],
      install_requires=[],
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      include_package_data=True,
      zip_safe=False)



import sys
import logging
import functools

import django
from django.conf import settings
from django.views.debug import SafeExceptionReporterFilter
from django.utils.module_loading import import_string

from .utils import get_django_notifier


logger = logging.getLogger('django')


class AirbrakeMiddleware:
  def __init__(self, get_response):
    self.get_response = get_response
    self._notifier = get_django_notifier()

  def __call__(self, request):
    response = self.get_response(request)
    return response

  def process_exception(self, request, exception):
    if not self._notifier:
      return

    notice = self._notifier.build_notice(exception)

    ctx = notice['context']
    ctx['url'] = request.build_absolute_uri()

    versions = notice['context'].get('versions', {})
    versions['django'] = django.get_version()
    ctx['versions'] = versions

    req_filter = get_exception_reporter_filter(request)
    notice['params'].update(dict(
      request_scheme=request.scheme,
      request_method=request.method,
      request_GET=request.GET,
      request_POST=req_filter.get_post_parameters(request),
      request_META=dict(request.META),
      request_FILES=request.FILES,
      request_COOKIES=request.COOKIES,
    ))

    if request.user.is_authenticated:
      user = request.user
      user_info = dict(
        username=user.username,
        email=user.email,
      )

      names = [user.first_name, user.last_name]
      names = [x for x in names if x]
      if len(names) > 0:
        user_info['name'] = ' '.join(names)

      notice['context']['user'] = user_info

    f = self._notifier.send_notice(notice)
    print(f.result())


@functools.lru_cache()
def get_default_exception_reporter_filter():
    # Instantiate the default filter for the first time and cache it.
    return import_string(settings.DEFAULT_EXCEPTION_REPORTER_FILTER)()


def get_exception_reporter_filter(request):
    default_filter = get_default_exception_reporter_filter()
    return getattr(request, 'exception_reporter_filter', default_filter)



import logging
from functools import lru_cache


@lru_cache()
def get_django_notifier():
  try:
    from django.conf import settings
  except ImportError:
    return

  if not hasattr(settings, 'AIRBRAKE'):
    if settings.DEBUG:
      logger.info('pybrake is not configured - set settings.AIRBRAKE')
      return

  from .notifier import Notifier
  return Notifier(**settings.AIRBRAKE)


@lru_cache()
def get_logger():
  logger = logging.getLogger('pybrake')
  fmter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

  sh = logging.StreamHandler()
  sh.setFormatter(fmter)
  logger.addHandler(sh)

  return logger
